import functools
import typing as ta

from .. import check
from .dispatch import Dispatcher
from .dispatch import get_impl_func_cls_set


# USE_EXTENSION = True
USE_EXTENSION = False


def function(func):  # noqa
    disp = Dispatcher()  # type: ignore
    disp.register(func, [object])

    func_name = getattr(func, '__name__', 'singledispatch function')

    def register(impl, cls=None):
        check.callable(impl)
        cls_col: ta.Iterable[type]
        if cls is None:
            cls_col = get_impl_func_cls_set(impl)
        else:
            cls_col = frozenset([cls])
        disp.register(impl, cls_col)
        return impl

    if not USE_EXTENSION:
        disp_dispatch = disp.dispatch

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            if not args:
                raise TypeError(f'{func_name} requires at least 1 positional argument')
            if (impl := disp_dispatch(type(args[0]))) is not None:
                return impl(*args, **kwargs)
            raise RuntimeError(f'No dispatch: {type(args[0])}')
        wrapper.register = register  # type: ignore
        wrapper.dispatch = disp.dispatch  # type: ignore

    else:
        from x.c._dispatch import function_wrapper  # noqa
        wrapper = function_wrapper(
            disp.dispatch,
            **{k: getattr(func, k) for k in functools.WRAPPER_ASSIGNMENTS if hasattr(func, k)},
            __wrapped__=func,
            register=register,
            dispatcher=disp,
        )

    return wrapper



"""
TODO:
 - backend abstr
 - streaming
"""
import functools
import json as _json
import typing as ta

from .. import lang


if ta.TYPE_CHECKING:
    import orjson as _orjson
    import ujson as _ujson

else:
    _orjson = lang.proxy_import('orjson')
    _ujson = lang.proxy_import('ujson')


##


dump = _json.dump
dumps = _json.dumps

detect_encoding = _json.detect_encoding

load = _json.load
loads = _json.loads


##


PRETTY_INDENT = 2

PRETTY_KWARGS: ta.Mapping[str, ta.Any] = dict(
    indent=PRETTY_INDENT,
)

dump_pretty: ta.Callable[..., bytes] = functools.partial(dump, **PRETTY_KWARGS)  # type: ignore
dumps_pretty: ta.Callable[..., str] = functools.partial(dumps, **PRETTY_KWARGS)


##


COMPACT_SEPARATORS = (',', ':')

COMPACT_KWARGS: ta.Mapping[str, ta.Any] = dict(
    indent=None,
    separators=COMPACT_SEPARATORS,
)

dump_compact: ta.Callable[..., bytes] = functools.partial(dump, **COMPACT_KWARGS)  # type: ignore
dumps_compact: ta.Callable[..., str] = functools.partial(dumps, **COMPACT_KWARGS)



from .cli import (  # noqa
    Inspect,
    Port,
    PsItem,
    cli_inspect,
    cli_ps,
    has_cli,
    parse_port,
)

from .compose import (  # noqa
    ComposeConfig,
    get_compose_port,
)

from .helpers import (  # noqa
    DOCKER_FOR_MAC_HOSTNAME,
    DOCKER_HOST_PLATFORM_KEY,
    get_docker_host_platform,
    is_likely_in_docker,
    timebomb_payload,
)

from .hub import (  # noqa
    HubRepoInfo,
    get_hub_repo_info,
)



import functools
import importlib.resources
import os.path
import typing as ta


class RelativeResource(ta.NamedTuple):
    name: str
    is_file: bool
    read_bytes: ta.Callable[[], bytes]


def get_relative_resources(
        path: str = '',
        *,
        globals: ta.Mapping[str, ta.Any] | None = None,  # noqa
        package: str | None = None,
        file: str | None = None,
) -> ta.Mapping[str, RelativeResource]:
    if globals is not None:
        if not package:
            package = globals.get('__package__')
        if not file:
            file = globals.get('__file__')

    #

    if os.sep in path:
        raise ValueError(path)  # noqa

    if not path.startswith('.'):
        path = '.' + path
    if set(path) - {'.'}:
        dot_pos = next(i for i in range(len(path)) if path[i] != '.')
        num_up = dot_pos - 1
        path_parts = path[dot_pos:].split('.')
    else:
        num_up = len(path) - 1
        path_parts = []

    #

    lst: list[RelativeResource] = []

    if package:
        pkg_parts = package.split('.')
        if num_up:
            pkg_parts = pkg_parts[:-num_up]
        anchor = '.'.join([*pkg_parts, *path_parts])
        for pf in importlib.resources.files(anchor).iterdir():
            lst.append(RelativeResource(
                name=pf.name,
                is_file=pf.is_file(),
                read_bytes=pf.read_bytes if pf.is_file() else None,  # type: ignore
            ))

    elif file:
        base_dir = os.path.dirname(file)
        dst_dir = os.path.join(
            base_dir,
            *(['..'] * num_up),
            *path_parts,
        )

        def _read_file(fp: str) -> bytes:
            with open(fp, 'rb') as f:
                return f.read()

        for ff in os.listdir(dst_dir):
            ff = os.path.join(dst_dir, ff)
            lst.append(RelativeResource(
                name=os.path.basename(ff),
                is_file=os.path.isfile(ff),
                read_bytes=functools.partial(_read_file, ff),
            ))

    else:
        raise RuntimeError('no package or file specified')

    return {r.name: r for r in lst}



import typing as ta

from .base import MarshalContext
from .base import Marshaler
from .base import TypeMapMarshalerFactory
from .base import TypeMapUnmarshalerFactory
from .base import UnmarshalContext
from .base import Unmarshaler
from .values import Value


##


PRIMITIVE_TYPES: tuple[type, ...] = (
    bool,
    int,
    float,
    str,
    bytes,
    type(None),
)


##


class PrimitiveMarshalerUnmarshaler(Marshaler, Unmarshaler):
    def marshal(self, ctx: MarshalContext, o: ta.Any) -> Value:
        if isinstance(o, PRIMITIVE_TYPES):
            return o  # type: ignore
        raise TypeError(o)

    def unmarshal(self, ctx: UnmarshalContext, v: Value) -> ta.Any:
        if isinstance(v, PRIMITIVE_TYPES):
            return v
        raise TypeError(v)


PRIMITIVE_MARSHALER_UNMARSHALER = PrimitiveMarshalerUnmarshaler()

PRIMITIVE_MARSHALER_FACTORY = TypeMapMarshalerFactory({  # noqa
    t: PRIMITIVE_MARSHALER_UNMARSHALER for t in PRIMITIVE_TYPES
})

PRIMITIVE_UNMARSHALER_FACTORY = TypeMapUnmarshalerFactory({  # noqa
    t: PRIMITIVE_MARSHALER_UNMARSHALER for t in PRIMITIVE_TYPES
})



import functools
import importlib
import os
import sys
import threading
import time
import traceback
import typing as ta


DEFAULT_TIMEOUT_S = 30

T = ta.TypeVar('T')
K = ta.TypeVar('K')
V = ta.TypeVar('V')


def assert_dicts_equal_ordered(l: ta.Mapping[K, V], r: ta.Mapping[K, V]) -> None:
    assert list(l.items()) == list(r.items())


def call_many_with_timeout(
        fns: ta.Iterable[ta.Callable[[], T]],
        timeout_s: float | None = None,
        timeout_exception: Exception = TimeoutError('Thread timeout'),
) -> list[T]:
    if timeout_s is None:
        timeout_s = DEFAULT_TIMEOUT_S

    fns = list(fns)
    missing = object()
    rets: list[ta.Any] = [missing] * len(fns)
    thread_exception: Exception | None = None

    def inner(fn, idx):
        try:
            nonlocal rets
            rets[idx] = fn()
        except Exception as e:
            nonlocal thread_exception
            thread_exception = e
            raise

    threads = [threading.Thread(target=inner, args=(fn, idx)) for idx, fn in enumerate(fns)]
    for thread in threads:
        thread.start()
    for thread in threads:
        thread.join(timeout_s)
    for thread in threads:
        if thread.is_alive():
            raise timeout_exception

    if thread_exception is not None:
        raise thread_exception
    for ret in rets:
        if ret is missing:
            raise ValueError

    return ta.cast(list[T], rets)


def run_with_timeout(
        *fns: ta.Callable[[], None],
        timeout_s: float | None = None,
        timeout_exception: Exception = TimeoutError('Thread timeout'),
) -> None:
    call_many_with_timeout(fns, timeout_s, timeout_exception)


def waitpid_with_timeout(
        pid: int,
        timeout_s: float | None = None,
        timeout_exception: Exception = TimeoutError('waitpid timeout'),
) -> int:
    if timeout_s is None:
        timeout_s = DEFAULT_TIMEOUT_S

    start_time = time.time()

    while True:
        wait_pid, status = os.waitpid(pid, os.WNOHANG)
        if wait_pid != 0:
            if wait_pid != pid:
                raise ValueError(f'{wait_pid} != {pid}')
            return status

        elapsed_time = time.time() - start_time
        if elapsed_time >= timeout_s:
            raise timeout_exception


def can_import(*args, **kwargs) -> bool:
    try:
        importlib.import_module(*args, **kwargs)
    except ImportError:
        return False
    else:
        return True


def xfail(fn):
    @functools.wraps(fn)
    def inner(*args, **kwargs):
        try:
            fn(*args, **kwargs)
        except Exception:  # noqa
            traceback.print_exc()

    return inner


def raise_in_thread(thr: threading.Thread, exc: BaseException | type[BaseException]) -> None:
    if sys.implementation.name != 'cpython':
        raise RuntimeError(sys.implementation.name)

    # https://github.com/python/cpython/blob/37ba7531a59a0a2b240a86f7e2adfb1b1cd8ac0c/Lib/test/test_threading.py#L182
    import ctypes as ct
    ct.pythonapi.PyThreadState_SetAsyncExc(ct.c_ulong(thr.ident), ct.py_object(exc))  # type: ignore



import os.path
import plistlib
import subprocess
import sys
import typing as ta

from .. import check
from .. import lang


if ta.TYPE_CHECKING:
    docker = lang.proxy_import('..docker')
else:
    from .. import docker


##


PYCHARM_HOSTED_ENV_VAR = 'PYCHARM_HOSTED'


def is_pycharm_hosted() -> bool:
    return PYCHARM_HOSTED_ENV_VAR in os.environ


##


PYCHARM_HOME = '/Applications/PyCharm.app'


def read_pycharm_info_plist() -> ta.Mapping[str, ta.Any] | None:
    plist_file = os.path.join(PYCHARM_HOME, 'Contents', 'Info.plist')
    if not os.path.isfile(plist_file):
        return None

    with open(plist_file, 'rb') as f:
        root = plistlib.load(f)

    return root


@lang.cached_function
def get_pycharm_version() -> str | None:
    plist = read_pycharm_info_plist()
    if plist is None:
        return None

    ver = check.non_empty_str(plist['CFBundleVersion'])
    check.state(ver.startswith('PY-'))
    return ver[3:]


##


def _import_pydevd_pycharm(*, version: str | None = None) -> ta.Any:
    if (
            'pydevd_pycharm' in sys.modules or
            (version is None and lang.can_import('pydevd_pycharm'))
    ):
        # Can't unload, nothing we can do
        import pydevd_pycharm  # noqa
        return pydevd_pycharm

    proc = subprocess.run([  # noqa
        sys.executable,
        '-m', 'pip',
        'show',
        'pydevd_pycharm',
    ], stdout=subprocess.PIPE)

    if not proc.returncode:
        info = {
            k: v.strip()
            for l in proc.stdout.decode().splitlines()
            if (s := l.strip())
            for k, _, v in [s.partition(':')]
        }

        installed_version = info['Version']
        if installed_version == version:
            import pydevd_pycharm  # noqa
            return pydevd_pycharm

        subprocess.check_call([
            sys.executable,
            '-m', 'pip',
            'uninstall', '-y',
            'pydevd_pycharm',
        ])

    subprocess.check_call([
        sys.executable,
        '-m', 'pip',
        'install',
        'pydevd_pycharm' + (f'=={version}' if version is not None else ''),
    ])

    import pydevd_pycharm  # noqa
    return pydevd_pycharm


def pycharm_remote_debugger_attach(
        host: str | None,
        port: int,
        *,
        version: str | None = None,
) -> None:
    # if version is None:
    #     version = get_pycharm_version()
    # check.non_empty_str(version)

    if host is None:
        if (
                sys.platform == 'linux' and
                docker.is_likely_in_docker() and
                docker.get_docker_host_platform() == 'darwin'
        ):
            host = docker.DOCKER_FOR_MAC_HOSTNAME
        else:
            host = 'localhost'

    if ta.TYPE_CHECKING:
        import pydevd_pycharm  # noqa
    else:
        pydevd_pycharm = _import_pydevd_pycharm(version=version)

    pydevd_pycharm.settrace(
        host,
        port=port,
        stdoutToServer=True,
        stderrToServer=True,
    )


# @omlish-manifest
_CLI_MODULE = {'$omdev.cli.types.CliModule': {
    'cmd_name': 'pycharm',
    'mod_name': __name__,
}}


if __name__ == '__main__':
    print(get_pycharm_version())



from .... import lang
from .base import Keyword
from .base import StrKeyword


##


class CoreKeyword(Keyword, lang.Abstract):
    pass


##


class Id(StrKeyword, CoreKeyword, lang.Final, tag='$id'):
    pass


class SchemaKeyword(StrKeyword, CoreKeyword, lang.Final, tag='$schema'):
    pass


class Ref(StrKeyword, CoreKeyword, lang.Final, tag='$ref'):
    pass



import contextlib


@contextlib.contextmanager
def attr_setting(obj, attr, val, *, default=None):  # noqa
    not_set = object()
    orig = getattr(obj, attr, not_set)
    try:
        setattr(obj, attr, val)
        if orig is not not_set:
            yield orig
        else:
            yield default
    finally:
        if orig is not_set:
            delattr(obj, attr)
        else:
            setattr(obj, attr, orig)



from .configs import (  # noqa
    configure_standard_logging,
)

from .formatters import (  # noqa
    ColorLogFormatter,
)

from .handlers import (  # noqa
    ListHandler,
)

from .utils import (  # noqa
    error_logging,
)


##


from ..lite.logs import (  # noqa
    TidLogFilter,
    JsonLogFormatter,

    STANDARD_LOG_FORMAT_PARTS,
    StandardLogFormatter,

    ProxyLogFilterer,
    ProxyLogHandler,

    StandardLogHandler,
)



import abc
import contextlib
import typing as ta
import weakref

from .. import c3
from .. import check
from .. import reflect as rfl


T = ta.TypeVar('T')


##


_IMPL_FUNC_CLS_SET_CACHE: ta.MutableMapping[ta.Callable, frozenset[type]] = weakref.WeakKeyDictionary()


def get_impl_func_cls_set(func: ta.Callable) -> frozenset[type]:
    with contextlib.suppress(KeyError):
        return _IMPL_FUNC_CLS_SET_CACHE[func]

    ann = getattr(func, '__annotations__', {})
    if not ann:
        raise TypeError(f'Invalid impl func: {func!r}')

    _, cls = next(iter(ta.get_type_hints(func).items()))
    rty = rfl.type_(cls)
    if isinstance(rty, rfl.Union):
        ret = frozenset(check.isinstance(arg, type) for arg in rty.args)
    else:
        ret = frozenset([check.isinstance(rty, type)])

    _IMPL_FUNC_CLS_SET_CACHE[func] = ret
    return ret


def find_impl(cls: type, registry: ta.Mapping[type, T]) -> T | None:
    mro = c3.compose_mro(cls, registry.keys())

    match: type | None = None
    for t in mro:
        if match is not None:
            # If *match* is an implicit ABC but there is another unrelated, equally matching implicit ABC, refuse the
            # temptation to guess.
            if (
                    t in registry
                    and t not in cls.__mro__
                    and match not in cls.__mro__
                    and not issubclass(match, t)
            ):
                raise RuntimeError(f'Ambiguous dispatch: {match} or {t}')
            break

        if t in registry:
            match = t

    if match is None:
        return None
    return registry.get(match)


##


class Dispatcher(ta.Generic[T]):
    def __init__(self) -> None:
        super().__init__()

        impls_by_arg_cls: dict[type, T] = {}
        self._impls_by_arg_cls = impls_by_arg_cls

        dispatch_cache: dict[ta.Any, T | None] = {}
        self._get_dispatch_cache = lambda: dispatch_cache

        def cache_remove(k, self_ref=weakref.ref(self)):
            if (ref_self := self_ref()) is not None:
                cache = ref_self._get_dispatch_cache()  # noqa
                with contextlib.suppress(KeyError):
                    del cache[k]

        cache_token: ta.Any = None
        self._get_cache_token = lambda: cache_token

        weakref_ref_ = weakref.ref

        def dispatch(cls: type) -> T | None:
            nonlocal cache_token

            if cache_token is not None and (current_token := abc.get_cache_token()) != cache_token:
                dispatch_cache.clear()
                cache_token = current_token

            cls_ref = weakref_ref_(cls)
            try:
                return dispatch_cache[cls_ref]
            except KeyError:
                pass
            del cls_ref

            impl: T | None
            try:
                impl = impls_by_arg_cls[cls]
            except KeyError:
                impl = find_impl(cls, impls_by_arg_cls)

            dispatch_cache[weakref_ref_(cls, cache_remove)] = impl
            return impl

        self.dispatch = dispatch

        def register(impl: T, cls_col: ta.Iterable[type]) -> T:
            nonlocal cache_token

            for cls in cls_col:
                impls_by_arg_cls[cls] = impl

                if cache_token is None and hasattr(cls, '__abstractmethods__'):
                    cache_token = abc.get_cache_token()

            dispatch_cache.clear()
            return impl

        self.register = register

    _get_cache_token: ta.Callable[[], int]
    _get_dispatch_cache: ta.Callable[[], ta.Any]

    def cache_size(self) -> int:
        return len(self._get_dispatch_cache())

    dispatch: ta.Callable[[type], T | None]

    register: ta.Callable[[T, ta.Iterable[type]], T]



"""
TODO:
 - @property dfs, bfs
"""
import functools
import typing as ta
import weakref

from .. import cached
from .. import check
from .. import collections as col
from .. import lang


T = ta.TypeVar('T')
NodeT = ta.TypeVar('NodeT')
NodeWalker = ta.Callable[[NodeT], ta.Iterable[NodeT]]
NodeGenerator = ta.Generator[NodeT, None, None]


class NodeError(ta.Generic[NodeT], Exception):
    def __init__(self, node: NodeT, msg: str, *args, **kwargs) -> None:
        super().__init__(msg, *args, **kwargs)  # noqa
        self._node = node

    @property
    def node(self) -> NodeT:
        return self._node


class DuplicateNodeError(NodeError[NodeT]):
    def __init__(self, node: NodeT, *args, **kwargs) -> None:
        super().__init__(node, f'Duplicate node: {node!r}', *args, **kwargs)


class UnknownNodeError(NodeError[NodeT]):
    def __init__(self, node: NodeT, *args, **kwargs) -> None:
        super().__init__(node, f'Unknown node: {node!r}', *args, **kwargs)


class BasicTreeAnalysis(ta.Generic[NodeT]):

    def __init__(
            self,
            root: NodeT,
            walker: NodeWalker[NodeT],
            *,
            identity: bool = False,
    ) -> None:
        super().__init__()

        self._root = root
        self._walker = walker
        self._identity = identity

        self._set_fac: ta.Callable[..., ta.MutableSet[NodeT]] = col.IdentitySet if identity else set
        self._dict_fac: ta.Callable[..., ta.MutableMapping[NodeT, ta.Any]] = col.IdentityKeyDict if identity else dict
        self._idx_seq_fac: ta.Callable[..., col.IndexedSeq[NodeT]] = functools.partial(col.IndexedSeq, identity=identity)  # type: ignore  # noqa

        def walk(cur: NodeT, parent: NodeT | None) -> None:
            check.not_none(cur)
            if cur in node_set:
                raise DuplicateNodeError(cur)

            nodes.append(cur)
            node_set.add(cur)
            if parent is None:
                check.state(cur is root)
            elif parent not in node_set:
                raise UnknownNodeError(parent)

            parents_by_node[cur] = parent

            children_by_node[cur] = children = list(walker(cur))
            child_sets_by_node[cur] = self._set_fac(children)
            for child in children:
                walk(child, cur)

        nodes: list[NodeT] = []
        node_set: ta.MutableSet[NodeT] = self._set_fac()
        children_by_node: ta.MutableMapping[NodeT | None, ta.Sequence[NodeT]] = self._dict_fac()
        child_sets_by_node: ta.MutableMapping[NodeT | None, ta.AbstractSet[NodeT]] = self._dict_fac()
        parents_by_node: ta.MutableMapping[NodeT, NodeT | None] = self._dict_fac()

        children_by_node[None] = [root]
        child_sets_by_node[None] = self._set_fac([root])

        walk(root, None)

        self._nodes = self._idx_seq_fac(nodes)
        self._node_set: ta.AbstractSet[NodeT] = node_set
        self._children_by_node: ta.Mapping[NodeT | None, col.IndexedSeq[NodeT]] = self._dict_fac(
            [(n, self._idx_seq_fac(cs)) for n, cs in children_by_node.items()])
        self._child_sets_by_node: ta.Mapping[NodeT | None, ta.AbstractSet[NodeT]] = child_sets_by_node
        self._parents_by_node: ta.Mapping[NodeT, NodeT | None] = parents_by_node

    @property
    def root(self) -> NodeT:
        return self._root

    @property
    def nodes(self) -> col.IndexedSeq[NodeT]:
        return self._nodes

    @property
    def walker(self) -> NodeWalker[NodeT]:
        return self._walker

    @property
    def identity(self) -> bool:
        return self._identity

    @property
    def node_set(self) -> ta.AbstractSet[NodeT]:
        return self._node_set

    @property
    def children_by_node(self) -> ta.Mapping[NodeT | None, col.IndexedSeq[NodeT]]:
        return self._children_by_node

    @property
    def child_sets_by_node(self) -> ta.Mapping[NodeT | None, ta.AbstractSet[NodeT]]:
        return self._child_sets_by_node

    @property
    def parents_by_node(self) -> ta.Mapping[NodeT, NodeT | None]:
        return self._parents_by_node

    @classmethod
    def from_parents(
            cls,
            src: ta.Mapping[NodeT, NodeT | None] | ta.Iterable[tuple[NodeT, NodeT | None]],
            *,
            identity: bool = False,
            **kwargs,
    ) -> 'BasicTreeAnalysis[NodeT]':
        pairs: ta.Sequence[tuple[NodeT, NodeT]]
        if isinstance(src, ta.Mapping):
            pairs = list(src.items())  # type: ignore
        elif isinstance(src, ta.Iterable):
            pairs = list(src)  # type: ignore
        else:
            raise TypeError(src)

        pairs = [(check.not_none(n), p) for n, p in pairs]

        root = check.single([n for n, p in pairs if p is None])  # noqa

        children_by_node: ta.MutableMapping[NodeT, ta.MutableSequence[NodeT]] = col.IdentityKeyDict() if identity else {}  # noqa
        for n, _ in pairs:
            children_by_node[n] = []
        for n, p in pairs:
            if p is not None:
                children_by_node[p].append(n)

        return cls(
            root,
            children_by_node.__getitem__,
            identity=identity,
            **kwargs,
        )

    @classmethod
    def from_children(
            cls,
            src: ta.Mapping[NodeT, ta.Iterable[NodeT]] | ta.Iterable[tuple[NodeT, ta.Iterable[NodeT]]],
            *,
            identity: bool = False,
            **kwargs,
    ) -> 'BasicTreeAnalysis[NodeT]':
        pairs: ta.Sequence[tuple[NodeT, ta.Sequence[NodeT]]]
        if isinstance(src, ta.Mapping):
            pairs = list(src.items())  # type: ignore
        elif isinstance(src, ta.Iterable):
            pairs = list(src)  # type: ignore
        else:
            raise TypeError(src)

        pairs = [(check.not_none(n), [check.not_none(c) for c in cs]) for n, cs in pairs]

        children_by_node: ta.MutableMapping[NodeT, ta.Sequence[NodeT]] = col.IdentityKeyDict() if identity else {}
        parents_by_node: ta.MutableMapping[NodeT, NodeT] = col.IdentityKeyDict() if identity else {}
        for n, cs in pairs:
            check.not_in(n, children_by_node)
            children_by_node[n] = cs
            for c in cs:
                check.not_in(c, parents_by_node)
                parents_by_node[c] = n

        e: ta.Any
        d: ta.Any
        if identity:
            e, d = id, col.make_map(((id(n), n) for n, _ in pairs), strict=True)
        else:
            e, d = lang.identity, lang.identity
        tsd = {e(n): {e(p)} for n, p in parents_by_node.items()}
        ts = list(col.mut_toposort(tsd))
        root = d(check.single(ts[0]))

        return cls(
            root,
            children_by_node.__getitem__,
            identity=identity,
            **kwargs,
        )

    @cached.property
    def _node_sets_by_type(self) -> ta.MutableMapping[type, ta.AbstractSet[NodeT]]:
        return weakref.WeakKeyDictionary()

    def get_node_type_set(self, ty: type[T]) -> ta.AbstractSet[T]:
        try:
            return self._node_sets_by_type[ty]  # type: ignore
        except KeyError:
            ret = self._node_sets_by_type[ty] = self._set_fac(n for n in self.nodes if isinstance(n, ty))
            return ret  # type: ignore

    def iter_ancestors(self, node: NodeT) -> NodeGenerator[NodeT]:
        cur: NodeT | None = node
        while True:
            cur = self.parents_by_node.get(cur)  # type: ignore
            if cur is None:
                break
            yield cur

    def get_lineage(self, node: NodeT) -> col.IndexedSeq[NodeT]:
        return self._idx_seq_fac(reversed([node, *self.iter_ancestors(node)]))

    def get_first_parent_of_type(self, node: NodeT, ty: type[T]) -> T | None:
        for cur in self.iter_ancestors(node):
            if isinstance(cur, ty):
                return cur
        return None

    @cached.property
    def depths_by_node(self) -> ta.Mapping[NodeT, int]:
        def rec(n, d):
            dct[n] = d
            for c in self._children_by_node[n]:
                rec(c, d + 1)
        dct: ta.MutableMapping[NodeT, int] = self._dict_fac()
        rec(self._root, 0)
        return dct



import operator
import typing as ta

from .... import check
from .... import collections as col
from .... import lang
from . import core  # noqa
from . import metadata  # noqa
from . import validation  # noqa
from .base import BooleanKeyword
from .base import Keyword
from .base import Keywords
from .base import KeywordsKeyword
from .base import NumberKeyword
from .base import StrKeyword
from .base import StrOrStrsKeyword
from .base import StrToKeywordsKeyword


KeywordT = ta.TypeVar('KeywordT', bound=Keyword)


##


KEYWORD_TYPES_BY_TAG: ta.Mapping[str, type[Keyword]] = col.make_map_by(  # noqa
    operator.attrgetter('tag'),
    (cls for cls in lang.deep_subclasses(Keyword) if not lang.is_abstract_class(cls)),
    strict=True,
)


def parse_keyword(cls: type[KeywordT], v: ta.Any) -> KeywordT:
    if issubclass(cls, BooleanKeyword):
        return cls(check.isinstance(v, bool))  # type: ignore

    elif issubclass(cls, NumberKeyword):
        return cls(check.isinstance(v, (int, float)))  # type: ignore

    elif issubclass(cls, StrKeyword):
        return cls(check.isinstance(v, str))  # type: ignore

    elif issubclass(cls, StrOrStrsKeyword):
        ss: str | ta.Sequence[str]
        if isinstance(v, str):
            ss = v
        elif isinstance(v, ta.Iterable):
            ss = col.seq_of(check.of_isinstance(str))(v)
        else:
            raise TypeError(v)
        return cls(ss)  # type: ignore

    elif issubclass(cls, KeywordsKeyword):
        return cls(parse_keywords(v))  # type: ignore

    elif issubclass(cls, StrToKeywordsKeyword):
        return cls({k: parse_keywords(mv) for k, mv in v.items()})  # type: ignore

    else:
        raise TypeError(cls)


def parse_keywords(dct: ta.Mapping[str, ta.Any]) -> Keywords:
    lst: list[Keyword] = []
    for k, v in dct.items():
        cls = KEYWORD_TYPES_BY_TAG[k]
        lst.append(parse_keyword(cls, v))
    return Keywords(lst)



import io
import itertools
import os
import signal
import sys
import threading
import time
import traceback
import typing as ta


##


def dump_threads(out: ta.IO) -> None:
    out.write('\n\n')

    thrs_by_tid = {t.ident: t for t in threading.enumerate()}

    for tid, fr in sys._current_frames().items():  # noqa
        try:
            thr = thrs_by_tid[tid]
        except KeyError:
            thr_rpr = repr(tid)
        else:
            thr_rpr = repr(thr)

        tb = traceback.format_stack(fr)

        out.write(f'{thr_rpr}\n')
        out.write('\n'.join(l.strip() for l in tb))
        out.write('\n\n')


def dump_threads_str() -> str:
    out = io.StringIO()
    dump_threads(out)
    return out.getvalue()


##


class StoppableThread:
    def __init__(
            self,
            fn: ta.Callable[[], None],
            interval_s: float,
            *,
            tick_immediately: bool = False,
            start: bool = False,
            **kwargs: ta.Any,
    ) -> None:
        super().__init__()
        self._fn = fn
        self._interval_s = interval_s
        self._tick_immediately = tick_immediately
        self._thread = threading.Thread(target=self._loop, **kwargs)
        self._stop_event = threading.Event()
        if start:
            self.start()

    @property
    def thread(self) -> threading.Thread:
        return self._thread

    @property
    def ident(self) -> int | None:
        return self._thread.ident

    def start(self) -> None:
        return self._thread.start()

    def stop_nowait(self) -> None:
        self._stop_event.set()

    def stop_wait(self, timeout: float | None = None) -> None:
        self.stop_nowait()
        self._thread.join(timeout)

    def _loop(self) -> None:
        if self._tick_immediately:
            self._fn()

        while True:
            self._stop_event.wait(self._interval_s)
            if self._stop_event.is_set():
                return

            self._fn()


##


_DEBUG_THREAD_COUNTER = itertools.count()


def create_thread_dump_thread(
        *,
        interval_s: float = 5.,
        out: ta.TextIO = sys.stderr,
        start: bool = False,
        nodaemon: bool = False,
) -> StoppableThread:
    def proc() -> None:
        try:
            out.write(dump_threads_str())
        except Exception as e:  # noqa
            out.write(repr(e) + '\n\n')

    return StoppableThread(
        proc,
        interval_s,
        daemon=not nodaemon,
        name=f'thread-dump-thread-{next(_DEBUG_THREAD_COUNTER)}',
        start=start,
    )


##


def create_suicide_thread(
        *,
        sig: int = signal.SIGKILL,
        interval_s: float = 1.,
        parent_thread: threading.Thread | None = None,
        start: bool = False,
) -> StoppableThread:
    """Kills process when parent_thread dies."""

    if parent_thread is None:
        parent_thread = threading.current_thread()

    def proc() -> None:
        if not parent_thread.is_alive():
            os.kill(os.getpid(), sig)

    return StoppableThread(
        proc,
        interval_s,
        name=f'suicide-thread-{next(_DEBUG_THREAD_COUNTER)}',
        start=start,
    )


##


def create_timebomb_thread(
        delay_s: float,
        *,
        sig: int = signal.SIGKILL,
        interval_s: float = 1.,
        start: bool = False,
) -> StoppableThread:
    def proc() -> None:
        if time.time() >= deadline:
            os.kill(os.getpid(), sig)

    deadline = time.time() + delay_s

    return StoppableThread(
        proc,
        interval_s,
        name=f'timebomb-thread-{next(_DEBUG_THREAD_COUNTER)}',
        start=start,
    )



"""
TODO:
 - cfg naming
 - adapters for dataclasses / namedtuples / user objects (as configured)
 - mro-merge ObjectMetadata
 - key ordering override - like slice, -1 means last
"""
import collections.abc
import typing as ta

from .. import check
from .. import dataclasses as dc
from .. import lang
from .. import reflect as rfl
from .base import MarshalContext
from .base import Marshaler
from .base import MarshalerFactory
from .base import UnmarshalContext
from .base import Unmarshaler
from .base import UnmarshalerFactory
from .naming import Naming
from .values import Value


##


@dc.dataclass(frozen=True, kw_only=True)
class FieldOptions:
    omit_if: ta.Callable[[ta.Any], bool] | None = None

    default: lang.Maybe[ta.Any] = dc.xfield(default=lang.empty(), check_type=lang.Maybe)

    embed: bool = False


DEFAULT_FIELD_OPTIONS = FieldOptions()
FIELD_OPTIONS_KWARGS: frozenset[str] = frozenset(dc.fields_dict(FieldOptions).keys())


@dc.dataclass(frozen=True, kw_only=True)
class FieldMetadata:
    name: str | None = None
    alts: ta.Iterable[str] | None = None

    options: FieldOptions = DEFAULT_FIELD_OPTIONS

    marshaler: Marshaler | None = None
    marshaler_factory: MarshalerFactory | None = None

    unmarshaler: Unmarshaler | None = None
    unmarshaler_factory: UnmarshalerFactory | None = None

    def update(self, **kwargs: ta.Any) -> 'FieldMetadata':
        okw = {k: v for k, v in kwargs.items() if k in FIELD_OPTIONS_KWARGS}
        mkw = {k: v for k, v in kwargs.items() if k not in FIELD_OPTIONS_KWARGS}
        return dc.replace(
            self,
            **(dict(options=dc.replace(self.options, **okw)) if okw else {}),
            **mkw,
        )


@dc.dataclass(frozen=True, kw_only=True)
class ObjectMetadata:
    field_naming: Naming | None = None

    unknown_field: str | None = None

    field_defaults: FieldMetadata = FieldMetadata()


##


@dc.dataclass(frozen=True, kw_only=True)
class FieldInfo:
    name: str
    type: ta.Any

    marshal_name: str
    unmarshal_names: ta.Sequence[str]

    metadata: FieldMetadata = FieldMetadata()

    options: FieldOptions = FieldOptions()


##


@dc.dataclass(frozen=True)
class ObjectMarshaler(Marshaler):
    fields: ta.Sequence[tuple[FieldInfo, Marshaler]]

    _: dc.KW_ONLY

    unknown_field: str | None = None

    def marshal(self, ctx: MarshalContext, o: ta.Any) -> Value:
        ret: dict[str, ta.Any] = {}
        for fi, m in self.fields:
            v = getattr(o, fi.name)

            if fi.options.omit_if is not None and fi.options.omit_if(v):
                continue

            mv = m.marshal(ctx, v)

            if fi.options.embed:
                for ek, ev in check.isinstance(mv, collections.abc.Mapping).items():
                    ret[fi.marshal_name + check.non_empty_str(ek)] = ev  # type: ignore

            else:
                ret[fi.marshal_name] = mv

        if self.unknown_field is not None:
            if (ukf := getattr(o, self.unknown_field)):
                if (dks := set(ret) & set(ukf)):
                    raise KeyError(f'Unknown field keys duplicate fields: {dks!r}')

            ret.update(ukf)  # FIXME: marshal?

        return ret


@dc.dataclass(frozen=True)
class SimpleObjectMarshalerFactory(MarshalerFactory):
    dct: ta.Mapping[type, ta.Sequence[FieldInfo]]

    _: dc.KW_ONLY

    unknown_field: str | None = None

    def guard(self, ctx: MarshalContext, rty: rfl.Type) -> bool:
        return isinstance(rty, type) and rty in self.dct

    def fn(self, ctx: MarshalContext, rty: rfl.Type) -> Marshaler:
        ty = check.isinstance(rty, type)

        flds = self.dct[ty]

        fields = [
            (fi, ctx.make(fi.type))
            for fi in flds
        ]

        return ObjectMarshaler(
            fields,
            unknown_field=self.unknown_field,
        )


##


@dc.dataclass(frozen=True)
class ObjectUnmarshaler(Unmarshaler):
    cls: type
    fields_by_unmarshal_name: ta.Mapping[str, tuple[FieldInfo, Unmarshaler]]

    _: dc.KW_ONLY

    unknown_field: str | None = None

    defaults: ta.Mapping[str, ta.Any] | None = None

    embeds: ta.Mapping[str, type] | None = None
    embeds_by_unmarshal_name: ta.Mapping[str, tuple[str, str]] | None = None

    def unmarshal(self, ctx: UnmarshalContext, v: Value) -> ta.Any:
        ma = check.isinstance(v, collections.abc.Mapping)

        u: ta.Any
        kw: dict[str, ta.Any] = {}
        ukf: dict[str, ta.Any] | None = None

        ekws: dict[str, dict[str, ta.Any]] = {en: {} for en in self.embeds or ()}

        if self.unknown_field is not None:
            kw[self.unknown_field] = ukf = {}

        for k, mv in ma.items():
            ks = check.isinstance(k, str)

            try:
                fi, u = self.fields_by_unmarshal_name[ks]

            except KeyError:
                if ukf is not None:
                    ukf[ks] = mv  # FIXME: unmarshal?
                    continue
                raise

            if self.embeds_by_unmarshal_name and (en := self.embeds_by_unmarshal_name.get(ks)):
                tkw, tk = ekws[en[0]], en[1]
            else:
                tkw, tk = kw, fi.name

            if tk in tkw:
                raise KeyError(f'Duplicate keys for field {tk!r}: {ks!r}')

            tkw[tk] = u.unmarshal(ctx, mv)

        for em, ecls in self.embeds.items() if self.embeds else ():
            ekw = ekws[em]
            ev = ecls(**ekw)
            kw[em] = ev

        if self.defaults:
            for dk, dv in self.defaults.items():
                kw.setdefault(dk, dv)

        return self.cls(**kw)


@dc.dataclass(frozen=True)
class SimpleObjectUnmarshalerFactory(UnmarshalerFactory):
    dct: ta.Mapping[type, ta.Sequence[FieldInfo]]

    _: dc.KW_ONLY

    unknown_field: str | None = None

    def guard(self, ctx: UnmarshalContext, rty: rfl.Type) -> bool:
        return isinstance(rty, type) and rty in self.dct

    def fn(self, ctx: UnmarshalContext, rty: rfl.Type) -> Unmarshaler:
        ty = check.isinstance(rty, type)

        flds = self.dct[ty]

        fields_by_unmarshal_name = {
            n: (fi, ctx.make(fi.type))
            for fi in flds
            for n in fi.unmarshal_names
        }

        defaults = {
            fi.name: fi.options.default.must()
            for fi in flds
            if fi.options.default.present
        }

        return ObjectUnmarshaler(
            ty,
            fields_by_unmarshal_name,
            unknown_field=self.unknown_field,
            defaults=defaults,
        )



# ruff: noqa: A002
# ruff: noqa: N802
# ruff: noqa: N815

import types
import typing as ta


##


Level: ta.TypeAlias = int


ExceptionInfo: ta.TypeAlias = tuple[type[BaseException], BaseException, types.TracebackType]


class LogRecord:
    name: str
    msg: str
    args: tuple
    levelname: str
    levelno: Level
    pathname: str
    filename: str
    module: str
    exc_info: ExceptionInfo | None
    exc_text: str | None
    stack_info: str | None
    lineno: int
    funcName: str
    created: float
    msecs: float
    relativeCreated: float
    thread: int
    threadName: str
    processName: str
    process: int


##


class Formatter(ta.Protocol):
    default_time_format: ta.ClassVar[str]
    default_msec_format: ta.ClassVar[str]

    def formatTime(self, record: LogRecord, datefmt: str | None = None) -> str: ...

    def formatException(self, ei: ExceptionInfo) -> str: ...

    def usesTime(self) -> bool: ...

    def formatMessage(self, record: LogRecord) -> str: ...

    def formatStack(self, stack_info: str) -> str: ...

    def format(self, record: LogRecord) -> str: ...


class BufferingFormatter(ta.Protocol):
    def formatHeader(self, records: ta.Sequence[LogRecord]) -> str: ...

    def formatFooter(self, records: ta.Sequence[LogRecord]) -> str: ...

    def format(self, records: ta.Sequence[LogRecord]) -> str: ...


##


class Filter(ta.Protocol):
    def filter(self, record: LogRecord) -> bool: ...


class Filterer(ta.Protocol):
    def addFilter(self, filter: Filter) -> None: ...

    def removeFilter(self, filter: Filter) -> None: ...

    def filter(self, record: LogRecord) -> bool: ...


##


class Handler(ta.Protocol):
    level: Level

    def get_name(self) -> str: ...

    def set_name(self, name: str) -> None: ...

    name: str

    def createLock(self) -> None: ...

    def acquire(self) -> None: ...

    def release(self) -> None: ...

    def setLevel(self, level: Level) -> None: ...

    def format(self, record: LogRecord) -> str: ...

    def emit(self, record: LogRecord) -> None: ...

    def handle(self, record: LogRecord) -> bool: ...

    def setFormatter(self, fmt: Formatter) -> None: ...

    def flush(self) -> None: ...

    def close(self) -> None: ...

    def handleError(self, record: LogRecord) -> None: ...


class Stream(ta.Protocol):
    def write(self, s: str) -> None: ...

    def flush(self) -> None: ...  # OPTIONAL METHOD

    def close(self) -> None: ...  # OPTIONAL METHOD


class StreamHandler(Handler):
    terminator: ta.ClassVar[str]

    stream: Stream

    def flush(self) -> None: ...

    def emit(self, record: LogRecord) -> None: ...

    def setStream(self, stream: Stream) -> None: ...


##


class Manager(ta.Protocol):
    root: 'Logger'

    disable: Level

    def getLogger(self, name: str) -> 'Logger': ...

    def setLoggerClass(self, klass: type['Logger']) -> None: ...

    def setLogRecordFactory(self, factory: ta.Callable[..., LogRecord]) -> None: ...  # UNREFERENCED?


##


Caller: ta.TypeAlias = tuple[
    str,  # filename
    int,  # lineno
    str,  # name
    str,  # formatted
]


class Logger(Filterer, ta.Protocol):
    name: str
    level: Level
    parent: ta.Optional['Logger']
    propagate: bool
    handlers: list[Handler]
    disabled: bool

    manager: Manager

    def setLevel(self, level: Level) -> None: ...

    def debug(self, msg: str, *args: ta.Any, **kwargs: ta.Any) -> None: ...
    def info(self, msg: str, *args: ta.Any, **kwargs: ta.Any) -> None: ...
    def warning(self, msg: str, *args: ta.Any, **kwargs: ta.Any) -> None: ...
    def warn(self, msg: str, *args: ta.Any, **kwargs: ta.Any) -> None: ...
    def error(self, msg: str, *args: ta.Any, **kwargs: ta.Any) -> None: ...
    def exception(self, msg: str, *args: ta.Any, **kwargs: ta.Any) -> None: ...
    def critical(self, msg: str, *args: ta.Any, **kwargs: ta.Any) -> None: ...
    def fatal(self, msg: str, *args: ta.Any, **kwargs: ta.Any) -> None: ...

    def log(self, level: Level, msg: str, *args: ta.Any, **kwargs: ta.Any) -> None: ...

    def findCaller(self, stack_info: bool = False, stacklevel: int = 1) -> Caller: ...

    def makeRecord(
        self,
        name,
        level,
        fn,
        lno,
        msg,
        args,
        exc_info,
        func: str | None = None,
        extra: ta.Mapping[str, ta.Any] | None = None,
        sinfo: str | None = None,
    ) -> LogRecord: ...

    def handle(self, record: LogRecord) -> None: ...

    def addHandler(self, hdlr: Handler) -> None: ...

    def removeHandler(self, hdlr: Handler) -> None: ...

    def hasHandlers(self) -> bool: ...

    def callHandlers(self, record: LogRecord) -> None: ...

    def getEffectiveLevel(self) -> Level: ...

    def isEnabledFor(self, level: Level) -> bool: ...

    def getChild(self, suffix: str) -> 'Logger': ...


##


class LoggerAdapter(ta.Protocol):
    logger: Logger
    extra: ta.Mapping[str, ta.Any]

    def process(self, msg: str, kwargs: dict[str, ta.Any]) -> tuple[str, dict[str, ta.Any]]: ...

    def debug(self, msg: str, *args: ta.Any, **kwargs: ta.Any) -> None: ...
    def info(self, msg: str, *args: ta.Any, **kwargs: ta.Any) -> None: ...
    def warning(self, msg: str, *args: ta.Any, **kwargs: ta.Any) -> None: ...
    def warn(self, msg: str, *args: ta.Any, **kwargs: ta.Any) -> None: ...
    def error(self, msg: str, *args: ta.Any, **kwargs: ta.Any) -> None: ...
    def exception(self, msg: str, *args: ta.Any, **kwargs: ta.Any) -> None: ...
    def critical(self, msg: str, *args: ta.Any, **kwargs: ta.Any) -> None: ...
    def fatal(self, msg: str, *args: ta.Any, **kwargs: ta.Any) -> None: ...

    def log(self, level: Level, msg: str, *args: ta.Any, **kwargs: ta.Any) -> None: ...

    def isEnabledFor(self, level: Level) -> bool: ...

    def setLevel(self, level: Level) -> None: ...

    def getEffectiveLevel(self) -> Level: ...

    def hasHandlers(self) -> bool: ...

    manager: Manager

    name: str



"""
Field:
    name: str | None = None
    type: Any = None
    default: Any | MISSING = MISSING
    default_factory: Any | MISSING = MISSING
    repr: bool = True
    hash: bool | None = None
    init: bool = True
    compare: bool = True
    metadata: Metadata | None = None
    kw_only: bool | MISSING = MISSING

    _field_type: Any = None


Params:
    init: bool = True
    repr: bool = True
    eq: bool = True
    order: bool = False
    unsafe_hash: bool = False
    frozen: bool = False

    match_args: bool = True
    kw_only: bool = False
    slots: bool = False
    weakref_slot: bool = False
"""
import dataclasses as dc
import typing as ta

from ... import lang
from .internals import PARAMS_ATTR
from .internals import Params
from .metadata import EMPTY_METADATA
from .metadata import METADATA_ATTR


##


@dc.dataclass(frozen=True, kw_only=True)
class FieldExtras(lang.Final):
    derive: ta.Callable[..., ta.Any] | None = None  # TODO
    coerce: bool | ta.Callable[[ta.Any], ta.Any] | None = None
    validate: ta.Callable[[ta.Any], bool] | None = None
    check_type: bool | type | None = None
    override: bool = False
    repr_fn: ta.Callable[[ta.Any], str | None] | None = None
    frozen: bool | None = None  # TODO


DEFAULT_FIELD_EXTRAS = FieldExtras()


def get_field_extras(f: dc.Field) -> FieldExtras:
    if not isinstance(f, dc.Field):
        raise TypeError(f)
    return f.metadata.get(FieldExtras, DEFAULT_FIELD_EXTRAS)


##


def get_params_cls(obj: ta.Any) -> type | None:
    if not isinstance(obj, type):
        obj = type(obj)
    for cur in obj.__mro__:
        if PARAMS_ATTR in cur.__dict__:
            return cur
    return None


def get_params(obj: ta.Any) -> Params:
    if not hasattr(obj, PARAMS_ATTR):
        raise TypeError(obj)
    return getattr(obj, PARAMS_ATTR)


##


@dc.dataclass(frozen=True, kw_only=True)
class ParamsExtras(lang.Final):
    reorder: bool = False
    cache_hash: bool = False
    generic_init: bool = False


DEFAULT_PARAMS_EXTRAS = ParamsExtras()


def get_params_extras(obj: ta.Any) -> ParamsExtras:
    if (pcls := get_params_cls(obj)) is None:
        raise TypeError(pcls)

    md = pcls.__dict__.get(METADATA_ATTR, EMPTY_METADATA)
    return md.get(ParamsExtras, DEFAULT_PARAMS_EXTRAS)


##


@dc.dataclass(frozen=True)
class MetaclassParams:
    confer: frozenset[str] = frozenset()


DEFAULT_METACLASS_PARAMS = MetaclassParams()


def get_metaclass_params(obj: ta.Any) -> MetaclassParams:
    if (pcls := get_params_cls(obj)) is None:
        raise TypeError(pcls)

    md = pcls.__dict__.get(METADATA_ATTR, EMPTY_METADATA)
    return md.get(MetaclassParams, DEFAULT_METACLASS_PARAMS)



"""
TODO:
 - visitor / transformer
 - uniform collection isinstance - items() for mappings, iter() for other
 - also check instance type in isinstance not just items lol
 - ta.Generic in mro causing trouble - omit? no longer 1:1
 - cache this shit, esp generic_mro shit
  - cache __hash__ in Generic/Union
"""
import dataclasses as dc
import types
import typing as ta


_NoneType = types.NoneType  # type: ignore

_NONE_TYPE_FROZENSET: frozenset['Type'] = frozenset([_NoneType])


_GenericAlias = ta._GenericAlias  # type: ignore  # noqa
_CallableGenericAlias = ta._CallableGenericAlias  # type: ignore  # noqa
_SpecialGenericAlias = ta._SpecialGenericAlias  # type: ignore  # noqa
_UnionGenericAlias = ta._UnionGenericAlias   # type: ignore  # noqa
_AnnotatedAlias = ta._AnnotatedAlias  # type: ignore  # noqa


##


@dc.dataclass(frozen=True)
class _Special:
    name: str
    alias: _SpecialGenericAlias  # type: ignore
    origin: type
    nparams: int


_KNOWN_SPECIALS = [
    _Special(
        v._name,  # noqa
        v,
        v.__origin__,
        v._nparams,  # noqa
    )
    for v in ta.__dict__.values()
    if isinstance(v, _SpecialGenericAlias)
]

_KNOWN_SPECIALS_BY_NAME = {s.name: s for s in _KNOWN_SPECIALS}
_KNOWN_SPECIALS_BY_ALIAS = {s.alias: s for s in _KNOWN_SPECIALS}
_KNOWN_SPECIALS_BY_ORIGIN = {s.origin: s for s in _KNOWN_SPECIALS}

_MAX_KNOWN_SPECIAL_TYPE_VARS = 16

_KNOWN_SPECIAL_TYPE_VARS = tuple(
    ta.TypeVar(f'_{i}')  # noqa
    for i in range(_MAX_KNOWN_SPECIAL_TYPE_VARS)
)


##


def get_params(obj: ta.Any) -> tuple[ta.TypeVar, ...]:
    if isinstance(obj, type):
        if issubclass(obj, ta.Generic):  # type: ignore
            return obj.__dict__.get('__parameters__', ())  # noqa

        if (ks := _KNOWN_SPECIALS_BY_ORIGIN.get(obj)) is not None:
            return _KNOWN_SPECIAL_TYPE_VARS[:ks.nparams]

    oty = type(obj)

    if (
            oty is _GenericAlias or
            oty is ta.GenericAlias  # type: ignore  # noqa
    ):
        return obj.__dict__.get('__parameters__', ())  # noqa

    if oty is _CallableGenericAlias:
        raise NotImplementedError('get_params not yet implemented for typing.Callable')

    raise TypeError(obj)


def is_union_type(cls: ta.Any) -> bool:
    if hasattr(ta, 'UnionType'):
        return ta.get_origin(cls) in {ta.Union, getattr(ta, 'UnionType')}

    else:
        return ta.get_origin(cls) in {ta.Union}


def get_orig_class(obj: ta.Any) -> ta.Any:
    return obj.__orig_class__  # noqa


##


Type: ta.TypeAlias = ta.Union[
    type,
    ta.TypeVar,
    'Union',
    'Generic',
    'NewType',
    'Annotated',
    'Any',
]


@dc.dataclass(frozen=True)
class Union:
    args: frozenset[Type]

    @property
    def is_optional(self) -> bool:
        return _NoneType in self.args

    def without_none(self) -> Type:
        if _NoneType not in self.args:
            return self
        rem = self.args - _NONE_TYPE_FROZENSET
        if len(rem) == 1:
            return next(iter(rem))
        return Union(rem)


@dc.dataclass(frozen=True)
class Generic:
    cls: type
    args: tuple[Type, ...]                                                # map[int, V] = (int, V) | map[T, T] = (T, T)

    params: tuple[ta.TypeVar, ...] = dc.field(compare=False, repr=False)  # map[int, V] = (_0, _1) | map[T, T] = (_0, _1)  # noqa
    # params2: tuple[ta.TypeVar, ...]                                     # map[int, V] = (V,)     | map[T, T] = (T,)

    obj: ta.Any = dc.field(compare=False, repr=False)

    # def __post_init__(self) -> None:
    #     if not isinstance(self.cls, type):
    #         raise TypeError(self.cls)

    def full_eq(self, other: 'Generic') -> bool:
        return (
            self.cls == other.cls and
            self.args == other.args and
            self.params == other.params and
            self.obj == other.obj
        )


@dc.dataclass(frozen=True)
class NewType:
    obj: ta.Any


@dc.dataclass(frozen=True)
class Annotated:
    ty: Type
    md: ta.Sequence[ta.Any]

    obj: ta.Any = dc.field(compare=False, repr=False)


class Any:
    pass


ANY = Any()


TYPES: tuple[type, ...] = (
    type,
    ta.TypeVar,
    Union,
    Generic,
    NewType,
    Annotated,
    Any,
)


##


def is_type(obj: ta.Any) -> bool:
    if isinstance(obj, (Union, Generic, ta.TypeVar, NewType, Any)):  # noqa
        return True

    oty = type(obj)

    return (
            oty is _UnionGenericAlias or oty is types.UnionType or  # noqa

            isinstance(obj, ta.NewType) or  # noqa

            (
                oty is _GenericAlias or
                oty is ta.GenericAlias or  # type: ignore  # noqa
                oty is _CallableGenericAlias
            ) or

            isinstance(obj, type) or

            isinstance(obj, _SpecialGenericAlias)
    )


def type_(obj: ta.Any) -> Type:
    if obj is ta.Any:
        return ANY

    if isinstance(obj, (Union, Generic, ta.TypeVar, NewType, Any)):  # noqa
        return obj

    oty = type(obj)

    if oty is _UnionGenericAlias or oty is types.UnionType:
        return Union(frozenset(type_(a) for a in ta.get_args(obj)))

    if isinstance(obj, ta.NewType):  # noqa
        return NewType(obj)

    if (
            oty is _GenericAlias or
            oty is ta.GenericAlias or  # type: ignore  # noqa
            oty is _CallableGenericAlias
    ):
        origin = ta.get_origin(obj)
        args = ta.get_args(obj)
        if oty is _CallableGenericAlias:
            p, r = args
            if p is Ellipsis or isinstance(p, ta.ParamSpec):
                raise TypeError(f'Callable argument not yet supported for {obj=}')
            args = (*p, r)
            params = _KNOWN_SPECIAL_TYPE_VARS[:len(args)]
        elif origin is ta.Generic:
            params = args
        else:
            params = get_params(origin)
        if len(args) != len(params):
            raise TypeError(f'Mismatched {args=} and {params=} for {obj=}')
        return Generic(
            origin,
            tuple(type_(a) for a in args),
            params,
            obj,
        )

    if isinstance(obj, type):
        if issubclass(obj, ta.Generic):  # type: ignore
            params = get_params(obj)
            if params:
                return Generic(
                    obj,
                    params,
                    params,
                    obj,
                )
        return obj

    if isinstance(obj, _SpecialGenericAlias):
        if (ks := _KNOWN_SPECIALS_BY_ALIAS.get(obj)) is not None:
            params = _KNOWN_SPECIAL_TYPE_VARS[:ks.nparams]
            return Generic(
                ks.origin,
                params,
                params,
                obj,
            )

    if isinstance(obj, _AnnotatedAlias):
        o = ta.get_args(obj)[0]
        return Annotated(type_(o), md=obj.__metadata__, obj=obj)

    raise TypeError(obj)



from .... import lang
from .base import BooleanKeyword
from .base import Keyword
from .base import KeywordsKeyword
from .base import NumberKeyword
from .base import StrOrStrsKeyword
from .base import StrToKeywordsKeyword


##


class ValidationKeyword(Keyword, lang.Abstract):
    pass


##


class Type(StrOrStrsKeyword, ValidationKeyword, lang.Final, tag='type'):
    pass


class Items(KeywordsKeyword, ValidationKeyword, lang.Final, tag='items'):
    pass


class Required(StrOrStrsKeyword, ValidationKeyword, lang.Final, tag='required'):
    pass


class Properties(StrToKeywordsKeyword, ValidationKeyword, lang.Final, tag='properties'):
    pass


##


class MaxItems(NumberKeyword, ValidationKeyword, lang.Final, tag='maxItems'):
    pass


class MinItems(NumberKeyword, ValidationKeyword, lang.Final, tag='minItems'):
    pass


class UniqueItems(BooleanKeyword, ValidationKeyword, lang.Final, tag='uniqueItems'):
    pass


#


class Maximum(NumberKeyword, ValidationKeyword, lang.Final, tag='maximum'):
    pass


class ExclusiveMaximum(NumberKeyword, ValidationKeyword, lang.Final, tag='exclusiveMaximum'):
    pass


class Minimum(NumberKeyword, ValidationKeyword, lang.Final, tag='minimum'):
    pass


class ExclusiveMinimum(NumberKeyword, ValidationKeyword, lang.Final, tag='exclusiveMinimum'):
    pass



"""
TODO:
 - redacted
 - strongly typed MarshalerFactory base class?
 - strongly typed Composite/Cached Marshaler/Unmarshaler factories - footgun
 - streaming? Start/EndObject, etc..
 - lang.Marker - class name, handle type[Foo]
  - can't disambiguate from str - can't coexist in bare union

See:
 - https://github.com/python-attrs/cattrs
 - https://github.com/jcrist/msgspec
 - https://github.com/Fatal1ty/mashumaro

cattrs:
 *

Jackson:
 - USE_ANNOTATIONS
 - AUTO_DETECT_CREATORS
 - AUTO_DETECT_FIELDS
 - AUTO_DETECT_GETTERS
 - AUTO_DETECT_IS_GETTERS
 - AUTO_DETECT_SETTERS
 - REQUIRE_SETTERS_FOR_GETTERS
 - USE_GETTERS_AS_SETTERS
 - INFER_CREATOR_FROM_CONSTRUCTOR_PROPERTIES
 - INFER_PROPERTY_MUTATORS
 - ALLOW_FINAL_FIELDS_AS_MUTATORS
 - ALLOW_VOID_VALUED_PROPERTIES
 - CAN_OVERRIDE_ACCESS_MODIFIERS
 - OVERRIDE_PUBLIC_ACCESS_MODIFIERS
 - SORT_PROPERTIES_ALPHABETICALLY
 - USE_WRAPPER_NAME_AS_PROPERTY_NAME
 - ACCEPT_CASE_INSENSITIVE_ENUMS
 - ACCEPT_CASE_INSENSITIVE_PROPERTIES
 - ACCEPT_CASE_INSENSITIVE_VALUES
 - ALLOW_EXPLICIT_PROPERTY_RENAMING
 - USE_STD_BEAN_NAMING
 - ALLOW_COERCION_OF_SCALARS
 - DEFAULT_VIEW_INCLUSION
 - IGNORE_DUPLICATE_MODULE_REGISTRATIONS
 - IGNORE_MERGE_FOR_UNMERGEABLE
 - USE_BASE_TYPE_AS_DEFAULT_IMPL
 - USE_STATIC_TYPING
 - BLOCK_UNSAFE_POLYMORPHIC_BASE_TYPES

https://github.com/yukinarit/pyserde
 - datatypes
  - typing.Union
  - typing.NewType for primitive types
  - typing.Any
  - typing.Literal
  - typing.Generic
  - typing.ClassVar
 - dataclasses.InitVar
 - Enum and IntEnum
 - pathlib.Path
 - decimal.Decimal
 - uuid.UUID
 - datetime.date, datetime.time, datetime.datetime
 - ipaddress
 - numpy types
 - Class Attributes
 - Field Attributes
 - Decorators
 - Type Check
 - Union Representation
 - Forward reference
 - PEP563 Postponed Evaluation of Annotations
 - PEP585 Type Hinting Generics In Standard Collections
 - PEP604 Allow writing union types as X | Y
 - PEP681 Data Class Transform
 - Case Conversion
 - Rename
 - Alias
 - Skip (de)serialization (skip, skip_if, skip_if_false, skip_if_default)
 - Custom field (de)serializer
 - Custom class (de)serializer
 - Custom global (de)serializer
 - Flatten
"""
import abc
import dataclasses as dc
import typing as ta

from .. import check
from .. import collections as col
from .. import lang
from .. import matchfns as mfs
from .. import reflect as rfl
from .exceptions import UnhandledTypeError
from .factories import RecursiveTypeFactory
from .factories import TypeCacheFactory
from .factories import TypeMapFactory
from .registries import Registry
from .registries import RegistryItem
from .utils import _Proxy
from .values import Value


##


class Marshaler(lang.Abstract):
    @abc.abstractmethod
    def marshal(self, ctx: 'MarshalContext', o: ta.Any) -> Value:
        raise NotImplementedError


class Unmarshaler(lang.Abstract):
    @abc.abstractmethod
    def unmarshal(self, ctx: 'UnmarshalContext', v: Value) -> ta.Any:
        raise NotImplementedError


MarshalerFactory: ta.TypeAlias = mfs.MatchFn[['MarshalContext', rfl.Type], Marshaler]
UnmarshalerFactory: ta.TypeAlias = mfs.MatchFn[['UnmarshalContext', rfl.Type], Unmarshaler]

MarshalerFactoryMatchClass: ta.TypeAlias = mfs.MatchFnClass[['MarshalContext', rfl.Type], Marshaler]
UnmarshalerFactoryMatchClass: ta.TypeAlias = mfs.MatchFnClass[['UnmarshalContext', rfl.Type], Unmarshaler]

#

TypeMapMarshalerFactory: ta.TypeAlias = TypeMapFactory['MarshalContext', Marshaler]
TypeMapUnmarshalerFactory: ta.TypeAlias = TypeMapFactory['UnmarshalContext', Unmarshaler]

TypeCacheMarshalerFactory: ta.TypeAlias = TypeCacheFactory['MarshalContext', Marshaler]
TypeCacheUnmarshalerFactory: ta.TypeAlias = TypeCacheFactory['UnmarshalContext', Unmarshaler]


##


@dc.dataclass(frozen=True)
class FuncMarshaler(Marshaler, lang.Final):
    fn: ta.Callable[['MarshalContext', ta.Any], Value]

    def marshal(self, ctx: 'MarshalContext', o: ta.Any) -> Value:
        return self.fn(ctx, o)


@dc.dataclass(frozen=True)
class FuncUnmarshaler(Unmarshaler, lang.Final):
    fn: ta.Callable[['UnmarshalContext', Value], ta.Any]

    def unmarshal(self, ctx: 'UnmarshalContext', v: Value) -> ta.Any:
        return self.fn(ctx, v)


##


class Option:
    pass


##


@dc.dataclass(frozen=True)
class BaseContext(lang.Abstract):
    registry: Registry
    options: col.TypeMap[Option] = col.TypeMap()


@dc.dataclass(frozen=True)
class MarshalContext(BaseContext, lang.Final):
    factory: MarshalerFactory | None = None

    def make(self, o: ta.Any) -> Marshaler:
        rty = rfl.type_(o)
        try:
            return check.not_none(self.factory)(self, rty)  # type: ignore
        except mfs.MatchGuardError:
            raise UnhandledTypeError(rty)  # noqa


@dc.dataclass(frozen=True)
class UnmarshalContext(BaseContext, lang.Final):
    factory: UnmarshalerFactory | None = None

    def make(self, o: ta.Any) -> Unmarshaler:
        rty = rfl.type_(o)
        try:
            return check.not_none(self.factory)(self, rty)  # type: ignore
        except mfs.MatchGuardError:
            raise UnhandledTypeError(rty)  # noqa


##


class _ProxyMarshaler(_Proxy[Marshaler], Marshaler):
    def marshal(self, ctx: MarshalContext, o: ta.Any) -> Value:
        return self._obj.marshal(ctx, o)


class RecursiveMarshalerFactory(RecursiveTypeFactory[MarshalContext, Marshaler], lang.Final):
    def __init__(self, f: MarshalerFactory) -> None:
        super().__init__(f, _ProxyMarshaler._new)  # noqa


class _ProxyUnmarshaler(_Proxy[Unmarshaler], Unmarshaler):
    def unmarshal(self, ctx: UnmarshalContext, v: Value) -> ta.Any:
        return self._obj.unmarshal(ctx, v)


class RecursiveUnmarshalerFactory(RecursiveTypeFactory[UnmarshalContext, Unmarshaler], lang.Final):
    def __init__(self, f: UnmarshalerFactory) -> None:
        super().__init__(f, _ProxyUnmarshaler._new)  # noqa


##


@dc.dataclass(frozen=True)
class SetType(RegistryItem, lang.Final):
    marshaler: Marshaler | None = None
    marshaler_factory: MarshalerFactory | None = None

    unmarshaler: Unmarshaler | None = None
    unmarshaler_factory: UnmarshalerFactory | None = None



def camel_case(name: str) -> str:
    return ''.join(map(str.capitalize, name.split('_')))  # noqa


def snake_case(name: str) -> str:
    uppers: list[int | None] = [i for i, c in enumerate(name) if c.isupper()]
    return '_'.join([name[l:r].lower() for l, r in zip([None, *uppers], [*uppers, None])]).strip('_')


def is_dunder(name: str) -> bool:
    return (
        name[:2] == name[-2:] == '__' and
        name[2:3] != '_' and
        name[-3:-2] != '_' and
        len(name) > 4
    )


def is_sunder(name: str) -> bool:
    return (
        name[0] == name[-1] == '_' and
        name[1:2] != '_' and
        name[-2:-1] != '_' and
        len(name) > 2
    )



# Copyright 2007 Pallets
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1.  Redistributions of source code must retain the above copyright notice, this list of conditions and the following
#     disclaimer.
#
# 2.  Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
#     following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3.  Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote
#     products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import datetime
import email.utils
import time

from .. import check


def _dt_as_utc(dt: datetime.datetime | None) -> datetime.datetime | None:
    if dt is None:
        return dt

    if dt.tzinfo is None:
        return dt.replace(tzinfo=datetime.UTC)
    elif dt.tzinfo != datetime.UTC:
        return dt.astimezone(datetime.UTC)

    return dt


def http_date(timestamp: datetime.datetime | datetime.date | float | time.struct_time | None = None) -> str:
    if isinstance(timestamp, datetime.date):
        if not isinstance(timestamp, datetime.datetime):
            # Assume plain date is midnight UTC.
            timestamp = datetime.datetime.combine(timestamp, datetime.time(), tzinfo=datetime.UTC)
        else:
            # Ensure datetime is timezone-aware.
            timestamp = _dt_as_utc(timestamp)

        return email.utils.format_datetime(check.not_none(timestamp), usegmt=True)

    if isinstance(timestamp, time.struct_time):
        timestamp = time.mktime(timestamp)

    return email.utils.formatdate(timestamp, usegmt=True)


def parse_date(value: str | None) -> datetime.datetime | None:
    if value is None:
        return None

    try:
        dt = email.utils.parsedate_to_datetime(value)
    except (TypeError, ValueError):
        return None

    if dt.tzinfo is None:
        return dt.replace(tzinfo=datetime.UTC)

    return dt



import random
import typing as ta

from .sorted import SortedCollection
from .sorted import SortedListDict


T = ta.TypeVar('T')
K = ta.TypeVar('K')
V = ta.TypeVar('V')


class SkipList(SortedCollection[T]):
    """https://gist.github.com/icejoywoo/3bf0c54983a725fa3917"""

    class _Node:
        __slots__ = [
            'value',
            'level',
            'next',
            'prev',
        ]

        next: list[ta.Optional['SkipList._Node']]
        prev: ta.Optional['SkipList._Node']

        def __init__(
                self,
                value: T,
                level: int,
        ) -> None:
            super().__init__()

            if level <= 0:
                raise TypeError('level must be > 0')

            self.value = value
            self.level = level
            self.next = [None] * level
            self.prev = None

        def __repr__(self) -> str:
            return f'{type(self).__name__}(value={self.value!r})'

    def __init__(
            self,
            *,
            max_height: int = 16,
            comparator: SortedCollection.Comparator[T] | None = None,
    ) -> None:
        super().__init__()

        if comparator is None:
            comparator = SortedCollection.default_comparator
        self._compare = comparator
        self._max_height = max_height
        self._head = SkipList._Node(None, self._max_height)
        self._height = 1
        self._head.next = [None] * self._max_height
        self._length = 0

    def __len__(self) -> int:
        return self._length

    def __iter__(self) -> ta.Iterator[T]:
        return iter(self.iter())

    def __contains__(self, value: T) -> bool:  # type: ignore
        return self.find(value) is not None

    def _random_level(self) -> int:
        result = 1
        while random.uniform(0, 1) < 0.5 and result < self._max_height:
            result += 1
        return result

    def add(self, value: T) -> bool:
        if value is None:
            raise TypeError(value)

        node = SkipList._Node(value, self._random_level())
        update = [None] * self._max_height  # noqa
        cur = self._head

        for i in range(self._height - 1, -1, -1):
            while cur.next[i] is not None and self._compare(value, cur.next[i].value) > 0:  # type: ignore
                cur = cur.next[i]  # type: ignore
            update[i] = cur  # type: ignore

        cur = cur.next[0]  # type: ignore
        if cur is not None:
            if self._compare(value, cur.value) == 0:  # type: ignore
                return False
            node.prev, cur.prev = cur.prev, node
        else:
            node.prev = update[0]  # type: ignore

        if node.level > self._height:
            for i in range(self._height, node.level):
                update[i] = self._head  # type: ignore
            self._height = node.level

        for i in range(node.level):
            cur = update[i]  # type: ignore
            node.next[i] = cur.next[i]  # noqa
            cur.next[i] = node  # noqa

        self._length += 1
        return True

    def _find(self, value: T) -> _Node | None:
        if value is None:
            raise TypeError(value)

        cur = self._head

        for i in range(self._height - 1, -1, -1):
            while cur.next[i] and self._compare(value, cur.next[i].value) > 0:  # type: ignore
                cur = cur.next[i]  # type: ignore

        return cur.next[0]

    def find(self, value: T) -> T | None:
        node = self._find(value)
        if node is None:
            return None
        if node is None or self._compare(value, node.value) != 0:  # type: ignore
            return None
        return node.value  # type: ignore

    def remove(self, value: T) -> bool:
        if value is None:
            raise TypeError(value)

        update = [None] * self._max_height  # noqa
        cur = self._head

        for i in range(self._height - 1, -1, -1):
            while cur.next[i] is not None and self._compare(value, cur.next[i].value) > 0:  # type: ignore
                cur = cur.next[i]  # type: ignore
            update[i] = cur  # type: ignore

        cur = cur.next[0]  # type: ignore
        if cur is None or self._compare(value, cur.value) != 0:  # type: ignore
            return False
        elif cur.next[0] is not None:
            cur.next[0].prev = cur.prev

        for i in range(self._height):
            if update[i].next[i] is not cur:  # type: ignore
                break
            update[i].next[i] = cur.next[i]  # type: ignore

        while self._height > 0 and self._head.next[self._height - 1] is None:
            self._height -= 1

        self._length -= 1
        return True

    def iter(self, base: T | None = None) -> ta.Iterable[T]:
        if base is not None:
            cur = self._find(base)
            while cur is not None and self._compare(base, cur.value) > 0:  # type: ignore
                cur = cur.next[0]
        else:
            cur = self._head.next[0]

        while cur is not None:
            yield cur.value  # type: ignore
            cur = cur.next[0]

    def riter(self, base: T | None = None) -> ta.Iterable[T]:
        if base is not None:
            cur = self._find(base)
            while cur is not self._head and self._compare(base, cur.value) < 0:  # type: ignore
                cur = cur.prev  # type: ignore
        else:
            cur = self._head.next[self._height - 1]
            while True:
                next = cur.next[cur.next.index(None) - 1 if None in cur.next else -1]  # type: ignore  # noqa
                if next is None:
                    break
                cur = next

        while cur is not self._head:
            yield cur.value  # type: ignore
            cur = cur.prev  # type: ignore


class SkipListDict(SortedListDict[K, V]):

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(SkipList(comparator=SortedListDict._item_comparator), *args, **kwargs)  # noqa



import functools
import operator
import types
import typing as ta


T = ta.TypeVar('T')
P = ta.ParamSpec('P')


##


def attr_property(n: str):
    return property(operator.attrgetter(n))


def item_property(n: str):
    return property(operator.itemgetter(n))


##


BUILTIN_METHOD_DESCRIPTORS = (classmethod, staticmethod)


class _MethodDescriptor:
    pass


def is_method_descriptor(obj: ta.Any) -> bool:
    return isinstance(obj, (*BUILTIN_METHOD_DESCRIPTORS, _MethodDescriptor))


def _has_method_descriptor(obj: ta.Any) -> bool:
    while True:
        if is_method_descriptor(obj):
            return True

        elif isinstance(obj, functools.partial):
            obj = obj.func

        else:
            try:
                obj = getattr(obj, '__wrapped__')
            except AttributeError:
                return False


def unwrap_method_descriptors(fn: ta.Callable) -> ta.Callable:
    while is_method_descriptor(fn):
        fn = fn.__func__  # type: ignore  # noqa
    return fn


##


def unwrap_func_with_partials(fn: ta.Callable) -> tuple[ta.Callable, list[functools.partial]]:
    ps = []
    while True:
        if is_method_descriptor(fn) or isinstance(fn, types.MethodType):
            fn = fn.__func__  # type: ignore

        elif hasattr(fn, '__wrapped__'):
            nxt = fn.__wrapped__
            if not callable(nxt):
                raise TypeError(nxt)
            if nxt is fn:
                raise TypeError(fn)
            fn = nxt

        # NOTE: __wrapped__ takes precedence - a partial might point to a bound Method when the important information is
        # still the unbound func. see _decorator_descriptor for an example of this.
        elif isinstance(fn, functools.partial):
            ps.append(fn)
            fn = fn.func

        else:
            break

    return fn, ps


def unwrap_func(fn: ta.Callable) -> ta.Callable:
    uw, _ = unwrap_func_with_partials(fn)
    return uw


##


def update_wrapper(
        wrapper: T,
        wrapped: ta.Any,
        assigned: ta.Iterable[str] = functools.WRAPPER_ASSIGNMENTS,
        updated: ta.Iterable[str] = functools.WRAPPER_UPDATES,
        *,
        filter: ta.Iterable[str] | None = None,  # noqa
        getattr: ta.Callable = getattr,  # noqa
        setattr: ta.Callable = setattr,  # noqa
) -> T:
    if filter:
        if isinstance(filter, str):
            filter = [filter]  # noqa
        assigned = tuple(a for a in assigned if a not in filter)
        updated = tuple(a for a in updated if a not in filter)

    for attr in assigned:
        try:
            value = getattr(wrapped, attr)
        except AttributeError:
            pass
        else:
            setattr(wrapper, attr, value)

    for attr in updated:
        getattr(wrapper, attr).update(getattr(wrapped, attr, {}))

    # Issue #17482: set __wrapped__ last so we don't inadvertently copy it from the wrapped function when updating
    # __dict__
    setattr(wrapper, '__wrapped__', wrapped)

    return wrapper


##


_DECORATOR_HANDLES_UNBOUND_METHODS = True


class _decorator_descriptor:  # noqa
    if not _DECORATOR_HANDLES_UNBOUND_METHODS:
        def __init__(self, wrapper, fn):
            self._wrapper, self._fn = wrapper, fn
            update_wrapper(self, fn, filter='__dict__')

        def __get__(self, instance, owner=None):
            return functools.update_wrapper(functools.partial(self._wrapper, fn := self._fn.__get__(instance, owner)), fn)  # noqa

    else:
        def __init__(self, wrapper, fn):
            self._wrapper, self._fn = wrapper, fn
            self._md = _has_method_descriptor(fn)
            update_wrapper(self, fn, filter='__dict__')

        def __get__(self, instance, owner=None):
            fn = self._fn.__get__(instance, owner)

            if self._md or instance is not None:
                @functools.wraps(fn)
                def inner(*args, **kwargs):
                    return self._wrapper(fn, *args, **kwargs)
                return inner

            else:
                @functools.wraps(fn)
                def outer(this, *args, **kwargs):
                    @functools.wraps(self._fn)
                    def inner(*args2, **kwargs2):
                        return fn(this, *args2, **kwargs2)
                    return self._wrapper(inner, *args, **kwargs)
                return outer

    def __repr__(self):
        return f'{self.__class__.__name__}<{self._wrapper}, {self._fn}>'

    def __call__(self, *args, **kwargs):
        return self._wrapper(self._fn, *args, **kwargs)


class _decorator:  # noqa
    def __init__(self, wrapper):
        self._wrapper = wrapper
        update_wrapper(self, wrapper, filter='__dict__')

    def __repr__(self):
        return f'{self.__class__.__name__}<{self._wrapper}>'

    def __call__(self, fn):
        return _decorator_descriptor(self._wrapper, fn)  # noqa


# FIXME:
# def decorator(
#         wrapper: ta.Callable[ta.Concatenate[ta.Any, P], T],  # FIXME: https://youtrack.jetbrains.com/issue/PY-72164  # noqa
# ) -> ta.Callable[[ta.Callable[P, T]], ta.Callable[P, T]]:
#     return _decorator(wrapper)


decorator = _decorator


##


class AccessForbiddenError(Exception):

    def __init__(self, name: str | None = None, *args: ta.Any, **kwargs: ta.Any) -> None:
        super().__init__(*((name,) if name is not None else ()), *args, **kwargs)  # noqa
        self.name = name


class AccessForbiddenDescriptor:

    def __init__(self, name: str | None = None) -> None:
        super().__init__()

        self._name = name

    def __set_name__(self, owner: type, name: str) -> None:
        if self._name is None:
            self._name = name
        elif name != self._name:
            raise NameError(name)

    def __get__(self, instance, owner=None):
        raise AccessForbiddenError(self._name)


def access_forbidden():  # noqa
    return AccessForbiddenDescriptor()


##


class _ClassOnly:
    def __init__(self, mth):
        if not isinstance(mth, classmethod):
            raise TypeError(f'must be classmethod: {mth}')
        super().__init__()
        self._mth = (mth,)
        functools.update_wrapper(self, mth)  # type: ignore

    @property
    def _mth(self):
        return self.__mth

    @_mth.setter
    def _mth(self, x):
        self.__mth = x

    def __get__(self, instance, owner=None):
        if instance is not None:
            raise TypeError(f'method must not be used on instance: {self._mth}')
        return self._mth[0].__get__(instance, owner)


def classonly(obj: T) -> T:  # noqa
    return _ClassOnly(obj)  # type: ignore



import dataclasses as dc
import threading
import typing as ta

from .. import check
from .. import lang
from .. import reflect as rfl


class RegistryItem(lang.Abstract):
    pass


RegistryItemT = ta.TypeVar('RegistryItemT', bound=RegistryItem)


@dc.dataclass(frozen=True)
class _TypeRegistry:
    rty: rfl.Type
    items: list[RegistryItem] = dc.field(default_factory=list)
    item_lists_by_ty: dict[type[RegistryItem], list[RegistryItem]] = dc.field(default_factory=dict)

    def add(self, *items: RegistryItem) -> None:
        for i in items:
            self.items.append(i)
            self.item_lists_by_ty.setdefault(type(i), []).append(i)


class Registry:
    def __init__(self) -> None:
        super().__init__()
        self._mtx = threading.Lock()
        self._dct: dict[rfl.Type, _TypeRegistry] = {}
        self._ps: ta.Sequence[Registry] = []

    def register(self, rty: rfl.Type, *items: RegistryItem) -> 'Registry':
        check.isinstance(rty, rfl.TYPES)
        with self._mtx:
            if (sr := self._dct.get(rty)) is None:
                sr = self._dct[rty] = _TypeRegistry(rty)
            sr.add(*items)
        return self

    def get(self, rty: rfl.Type) -> ta.Sequence[RegistryItem]:
        check.isinstance(rty, rfl.TYPES)
        try:
            return self._dct[rty].items
        except KeyError:
            return ()

    def get_of(self, rty: rfl.Type, item_ty: type[RegistryItemT]) -> ta.Sequence[RegistryItemT]:
        check.isinstance(rty, rfl.TYPES)
        try:
            sr = self._dct[rty]
        except KeyError:
            return ()
        return sr.item_lists_by_ty.get(item_ty, ())  # type: ignore



import enum
import re
import typing as ta

from . import lang


ESC = '\x1b'
BEL = '\x07'

DCS = ESC + 'P'
CSI = ESC + '['
ST = ESC + '\\'
OSC = ESC + ']'
RIS = ESC + 'c'


def set_title(title: str) -> str:
    return OSC + '0;' + title + BEL


def strip_ansi_codes(s: str) -> str:
    return re.sub(r'\033\[([0-9]+)(;[0-9]+)*m', '', s)


class ControlSequence:

    def __init__(self, fn: ta.Callable[..., str], desc: str) -> None:
        super().__init__()
        self._fn = fn
        self._desc = desc

    @property
    def description(self) -> str:
        return self._desc

    def __call__(self, *args, **kwargs) -> str:
        return self._fn(*args, **kwargs)


CUU = ControlSequence(lambda n: CSI + str(n) + 'A', 'Cursor Up')
CUD = ControlSequence(lambda n: CSI + str(n) + 'B', 'Cursor Down')
CUF = ControlSequence(lambda n: CSI + str(n) + 'C', 'Cursor Forward')
CUB = ControlSequence(lambda n: CSI + str(n) + 'D', 'Cursor Back')

CNL = ControlSequence(lambda n: CSI + str(n) + 'E', 'Cursor Next Line')
CPL = ControlSequence(lambda n: CSI + str(n) + 'F', 'Cursor Previous Line')
CHA = ControlSequence(lambda n: CSI + str(n) + 'G', 'Cursor Horizontal Line Absolute')

CUP = ControlSequence(lambda n, m: CSI + str(n) + ';' + str(m) + 'H', 'Cursor Position')
HVP = ControlSequence(lambda n, m: CSI + str(n) + ';' + str(m) + 'f', 'Horizontal Vertical Position')


def _str_val(val: ta.Any) -> str:
    if isinstance(val, enum.Enum):
        return str(val.value)
    else:
        return str(val)


class EDs(enum.Enum):
    FROM_CURSOR_TO_END = 0
    FROM_CURSOR_TO_BEGINNING = 1
    ALL = 2
    ALL_AND_SCROLLBACK = 3


ED = ControlSequence(lambda n: CSI + _str_val(n) + 'J', 'Erase in Display')


class ELs(enum.Enum):
    FROM_CURSOR_TO_END = 0
    FROM_CURSOR_TO_BEGINNING = 1
    ALL = 2


EL = ControlSequence(lambda n: CSI + _str_val(n) + 'K', 'Erase in Line')

SU = ControlSequence(lambda n: CSI + str(n) + 'S', 'Scroll Up')
SD = ControlSequence(lambda n: CSI + str(n) + 'T', 'Scroll Down')

DSR = ControlSequence(lambda: CSI + '6n', 'Device Status Report')

SCP = ControlSequence(lambda: CSI + 's', 'Save Cursor Position')
RCP = ControlSequence(lambda: CSI + 'u', 'Restore Cursor Position')

SHOW_CURSOR = ControlSequence(lambda: CSI + '?25h', 'Show Cursor')
HIDE_CURSOR = ControlSequence(lambda: CSI + '?25l', 'Hide Cursor')

SGR = ControlSequence(lambda n: CSI + _str_val(n) + 'm', 'Select Graphic Rendition')


class SGRs(lang.Namespace, lang.Final):
    RESET = 0
    NORMAL_COLOR_AND_INTENSITY = 22

    class FONT(enum.Enum):
        BOLD = 1
        FAINT = 2
        ITALIC = 3
        UNTERLINE = 4

        SLOW_BLINK = 5
        RAPID_BLINK = 6

        REVERSE_VIDEO = 7

        PRIMARY_FONT = 10
        ITALIC_OFF = 23
        UNDERLINE_OFF = 24
        BLINK_OFF = 25
        INVERSE_OFF = 27

        FRAMED = 51
        ENCIRCLED = 52
        OVERLINED = 53
        NOT_FRAMED_OR_ENCIRCLED = 54
        NOT_OVERLINED = 55

    class FG(enum.Enum):
        BLACK = 30
        RED = 31
        GREEN = 32
        YELLOW = 33
        BLUE = 34
        MAGENTA = 35
        CYAN = 36
        WHITE = 37

        BRIGHT_BLACK = 90
        BRIGHT_RED = 91
        BRIGHT_GREEN = 92
        BRIGHT_YELLOW = 93
        BRIGHT_BLUE = 94
        BRIGHT_MAGENTA = 95
        BRIGHT_CYAN = 96
        BRIGHT_WHITE = 97

    class BG(enum.Enum):
        BLACK = 40
        RED = 41
        GREEN = 42
        YELLOW = 43
        BLUE = 44
        MAGENTA = 45
        CYAN = 46
        WHITE = 47

        BRIGHT_BLACK = 100
        BRIGHT_RED = 101
        BRIGHT_GREEN = 102
        BRIGHT_YELLOW = 103
        BRIGHT_BLUE = 104
        BRIGHT_MAGENTA = 105
        BRIGHT_CYAN = 106
        BRIGHT_WHITE = 107


def _clamp_ofs(v: int, hi: int, ofs: int) -> str:
    if v < 0 or v > hi:
        raise ValueError(v)
    return str(v + ofs)


FG8 = ControlSequence(
    lambda n: CSI + '38;5;' + str(n) + 'm',
    '8-Bit Foreground Color')
FG8_STANDARD = ControlSequence(
    lambda n: CSI + '38;5;' + _clamp_ofs(n, 8, 0) + 'm',
    '8-Bit Foreground Color (Standard)')
FG8_HIGH_INTENSITY = ControlSequence(
    lambda n: CSI + '38;5;' + _clamp_ofs(n, 8, 8) + 'm',
    '8-Bit Foreground Color (High Intensity)')
FG8_216 = ControlSequence(
    lambda n: CSI + '38;5;' + _clamp_ofs(n, 216, 16) + 'm',
    '8-Bit Foreground Color (High Intensity)')
FG8_GRAYSCALE = ControlSequence(
    lambda n: CSI + '38;5;' + _clamp_ofs(n, 24, 232) + 'm',
    '8-Bit Foreground Color (Grayscale)')
FG8_RGB = ControlSequence(
    lambda r, g, b: CSI + '38;5;' + str(36 * r + 6 * g + b) + 'm',
    '8-Bit Foreground Color (RGB)')

BG8 = ControlSequence(
    lambda n: CSI + '48;5;' + str(n) + 'm',
    '8-Bit Background Color')
BG8_STANDARD = ControlSequence(
    lambda n: CSI + '48;5;' + _clamp_ofs(n, 8, 0) + 'm',
    '8-Bit Background Color (Standard)')
BG8_HIGH_INTENSITY = ControlSequence(
    lambda n: CSI + '48;5;' + _clamp_ofs(n, 8, 8) + 'm',
    '8-Bit Background Color (High Intensity)')
BG8_216 = ControlSequence(
    lambda n: CSI + '48;5;' + _clamp_ofs(n, 216, 16) + 'm',
    '8-Bit Background Color (High Intensity)')
BG8_GRAYSCALE = ControlSequence(
    lambda n: CSI + '48;5;' + _clamp_ofs(n, 24, 232) + 'm',
    '8-Bit Background Color (Grayscale)')
BG8_RGB = ControlSequence(
    lambda r, g, b: CSI + '48;5;' + str(36 * r + 6 * g + b) + 'm',
    '8-Bit Background Color (RGB)')

FG24_RGB = ControlSequence(
    lambda r, g, b: CSI + '38;2;' + str(r) + ';' + str(g) + ';' + str(b) + 'm',
    '24-Bit Foreground Color (RGB)')
BG24_RGB = ControlSequence(
    lambda r, g, b: CSI + '48;2;' + str(r) + ';' + str(g) + ';' + str(b) + 'm',
    '24-Bit Background Color (RGB)')


def main() -> None:
    import sys

    sys.stdout.write(SGR(SGRs.RESET))
    sys.stdout.write(BG8(15) + ' ')
    for i in range(20):
        sys.stdout.write(BG8(i) + ' ')
    sys.stdout.write('\n')
    for i in range(256):
        if i % 12 == 0:
            sys.stdout.write('\n')
        sys.stdout.write(BG8(i + 16) + ' ')
    sys.stdout.write(SGR(SGRs.RESET) + '\n')


if __name__ == '__main__':
    main()



from .types import (  # noqa
    ANY,
    Annotated,
    Any,
    Generic,
    NewType,
    TYPES,
    Type,
    Union,
    get_orig_class,
    get_params,
    is_type,
    is_union_type,
    type_,
)

from .ops import (  # noqa
    get_concrete_type,
    get_underlying,
    strip_annotations,
    strip_objs,
    to_annotation,
    types_equivalent,
)

from .isinstance import (  # noqa
    KNOWN_ISINSTANCE_GENERICS,
    isinstance_of,
)

from .subst import (  # noqa
    ALIAS_UPDATING_GENERIC_SUBSTITUTION,
    DEFAULT_GENERIC_SUBSTITUTION,
    GenericSubstitution,
    generic_mro,
    get_generic_bases,
    get_type_var_replacements,
    replace_type_vars,
)



import abc
import functools
import typing as ta

from .abstract import Abstract
from .abstract import is_abstract


##


class FinalError(TypeError):
    def __init__(self, _type: type) -> None:
        super().__init__()

        self._type = _type

    def __repr__(self) -> str:
        return f'{type(self).__name__}({self._type})'


class Final(Abstract):
    __slots__ = ()

    def __init_subclass__(cls, **kwargs: ta.Any) -> None:
        super().__init_subclass__(**kwargs)

        abstracts: set[ta.Any] = set()
        for base in cls.__bases__:
            if base is Abstract:
                raise FinalError(base)
            elif base is Final:
                continue
            elif Final in base.__mro__:
                raise FinalError(base)
            else:
                abstracts.update(getattr(base, '__abstractmethods__', []))

        for a in abstracts:
            try:
                v = cls.__dict__[a]
            except KeyError:
                raise FinalError(a) from None
            if is_abstract(v):
                raise FinalError(a)


##


class SealedError(TypeError):
    def __init__(self, _type) -> None:
        super().__init__()

        self._type = _type

    def __repr__(self) -> str:
        return f'{type(self).__name__}({self._type})'


class Sealed:
    __slots__ = ()

    def __init_subclass__(cls, **kwargs: ta.Any) -> None:
        for base in cls.__bases__:
            if base is not Abstract:
                if Sealed in base.__bases__:
                    if cls.__module__ != base.__module__:
                        raise SealedError(base)
        super().__init_subclass__(**kwargs)


class PackageSealed:
    __slots__ = ()

    def __init_subclass__(cls, **kwargs: ta.Any) -> None:
        for base in cls.__bases__:
            if base is not Abstract:
                if PackageSealed in base.__bases__:
                    pfx = base.__module__.split('.')[:-1]
                    if cls.__module__.split('.')[:len(pfx)] != pfx:
                        raise SealedError(base)
        super().__init_subclass__(**kwargs)


##


class NotInstantiable(Abstract):
    __slots__ = ()

    def __new__(cls, *args, **kwargs):  # noqa
        raise TypeError


class NotPicklable:
    __slots__ = ()

    @ta.final
    def __reduce__(self) -> ta.NoReturn:
        raise TypeError

    @ta.final
    def __reduce_ex__(self, protocol) -> ta.NoReturn:
        raise TypeError

    @ta.final
    def __getstate__(self) -> ta.NoReturn:
        raise TypeError

    @ta.final
    def __setstate__(self, state) -> ta.NoReturn:
        raise TypeError


##


class NoBool:
    def __bool__(self) -> bool:
        raise TypeError


class _NoBoolDescriptor:
    def __init__(self, fn, instance=None, owner=None) -> None:
        super().__init__()
        self._fn = fn
        self._instance = instance
        self._owner = owner
        functools.update_wrapper(self, fn)

    def __bool__(self) -> bool:
        raise TypeError

    def __get__(self, instance, owner=None):
        if instance is self._instance and owner is self._owner:
            return self
        return _NoBoolDescriptor(self._fn.__get__(instance, owner), instance, owner)

    def __call__(self, *args, **kwargs):
        return self._fn(*args, **kwargs)


def no_bool(fn):  # noqa
    return _NoBoolDescriptor(fn)


##


SENSITIVE_ATTR = '__sensitive__'


class Sensitive:
    __sensitive__ = True


class _AnySensitiveMeta(abc.ABCMeta):
    @classmethod
    def __instancecheck__(cls, instance: object) -> bool:
        return hasattr(instance, SENSITIVE_ATTR) or super().__instancecheck__(AnySensitive, instance)

    @classmethod
    def __subclasscheck__(cls, subclass: type) -> bool:
        return hasattr(subclass, SENSITIVE_ATTR) or super().__subclasscheck__(AnySensitive, subclass)


class AnySensitive(NotInstantiable, Final, metaclass=_AnySensitiveMeta):
    pass
